package klaue.minecraftimagemap;

import java.awt.Color;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;

/**
 * Most code here is from
 * http://stackoverflow.com/questions/5940188/how-to-convert-a-24-bit-png-to-3-bit-png-using-floydsteinberg-dithering
 * floydSteinbergDithering was slightly changed to allow a given
 * color array (2nd parameter) and copy the given image instead of changing it,
 * C3 was changed to add a color CTOR
 * function "noDithering" was added
 * 
 * @author klaue
 * 
 */
public class ColorIndexer {
	public static BufferedImage floydSteinbergDithering(BufferedImage img, Color[] colors) {

		C3[] palette = new C3[colors.length];
		for (int i = 0; i < colors.length; ++i) {
			palette[i] = new C3(colors[i]);
		}

		int w = img.getWidth();
		int h = img.getHeight();

		ColorModel cm = img.getColorModel();
		BufferedImage newImg = new BufferedImage(cm, img.copyData(null),
				cm.isAlphaPremultiplied(), null);

		C3[][] d = new C3[h][w];

		for (int y = 0; y < h; y++)
			for (int x = 0; x < w; x++)
				d[y][x] = new C3(newImg.getRGB(x, y));

		for (int y = 0; y < newImg.getHeight(); y++) {
			for (int x = 0; x < newImg.getWidth(); x++) {

				C3 oldColor = d[y][x];
				C3 newColor = findClosestPaletteColor(oldColor, palette);
				newImg.setRGB(x, y, newColor.toColor().getRGB());

				C3 err = oldColor.sub(newColor);

				if (x + 1 < w)
					d[y][x + 1] = d[y][x + 1].add(err.mul(7. / 16));
				if (x - 1 >= 0 && y + 1 < h)
					d[y + 1][x - 1] = d[y + 1][x - 1].add(err.mul(3. / 16));
				if (y + 1 < h)
					d[y + 1][x] = d[y + 1][x].add(err.mul(5. / 16));
				if (x + 1 < w && y + 1 < h)
					d[y + 1][x + 1] = d[y + 1][x + 1].add(err.mul(1. / 16));
			}
		}

		return newImg;
	}
	
	public static BufferedImage noDithering(BufferedImage img, Color[] colors) {
		// this would not have to use C3 at all, but meh,  use existing code
		C3[] palette = new C3[colors.length];
		for (int i = 0; i < colors.length; ++i) {
			palette[i] = new C3(colors[i]);
		}

		int w = img.getWidth();
		int h = img.getHeight();

		ColorModel cm = img.getColorModel();
		BufferedImage newImg = new BufferedImage(cm, img.copyData(null),
				cm.isAlphaPremultiplied(), null);

		C3[][] d = new C3[h][w];

		for (int y = 0; y < h; y++)
			for (int x = 0; x < w; x++)
				d[y][x] = new C3(newImg.getRGB(x, y));

		for (int y = 0; y < newImg.getHeight(); y++) {
			for (int x = 0; x < newImg.getWidth(); x++) {

				C3 oldColor = d[y][x];
				C3 newColor = findClosestPaletteColor(oldColor, palette);
				newImg.setRGB(x, y, newColor.toColor().getRGB());
			}
		}

		return newImg;
	}

	private static C3 findClosestPaletteColor(C3 c, C3[] palette) {
		C3 closest = palette[0];

		for (C3 n : palette)
			if (n.diff(c) < closest.diff(c))
				closest = n;

		return closest;
	}

	static class C3 {
		int r, g, b;

		public C3(int c) {
			this(new Color(c));
		}

		public C3(int r, int g, int b) {
			this.r = r;
			this.g = g;
			this.b = b;
		}
		
		public C3(Color color) {
			this.r = color.getRed();
			this.g = color.getGreen();
			this.b = color.getBlue();
		}

		public C3 add(C3 o) {
			return new C3(r + o.r, g + o.g, b + o.b);
		}

		public C3 sub(C3 o) {
			return new C3(r - o.r, g - o.g, b - o.b);
		}

		public C3 mul(double d) {
			return new C3((int) (d * r), (int) (d * g), (int) (d * b));
		}

		public int toRGB() {
			return toColor().getRGB();
		}

		public Color toColor() {
			return new Color(clamp(r), clamp(g), clamp(b));
		}

		public int clamp(int c) {
			return Math.max(0, Math.min(255, c));
		}

		public int diff(C3 o) {
			int Rdiff = o.r - this.r;
			int Gdiff = o.g - this.g;
			int Bdiff = o.b - this.b;
			int distanceSquared = Rdiff * Rdiff + Gdiff * Gdiff + Bdiff * Bdiff;
			return distanceSquared;
		}
	}
}
