package klaue.minecraftimagemap;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import javax.imageio.ImageIO;

import org.jnbt.ByteArrayTag;
import org.jnbt.ByteTag;
import org.jnbt.CompoundTag;
import org.jnbt.IntTag;
import org.jnbt.NBTOutputStream;
import org.jnbt.ShortTag;
import org.jnbt.Tag;

public class Exporter {
	public static HashMap<Integer, BufferedImage> minecraftBlocks = null;
	
	public static void exportAsPNG(BufferedImage image, File file) throws IOException {
		BufferedImage saveImg = getSquareImg(image);
		ImageIO.write(saveImg, "png", file);
	}
	
	public static void exportAsBlocks(BufferedImage image, File file, MinecraftMapColor[] minecraftMapColors) throws IOException {
		BufferedImage squareImg = getSquareImg(image);
		int width = squareImg.getWidth();
		int height = squareImg.getHeight();
		
		// each pixel of the source image will be 16x16 pixels in MC blocks. Also, there will be gridlines
		int newW = (width*16) + (width-1);
		int newH = (height*16) + (height-1);
		BufferedImage saveImg = new BufferedImage(newW, newH, BufferedImage.TYPE_INT_ARGB);
		Graphics2D g = saveImg.createGraphics();
		g.setColor(new Color(255, 255, 255, 255));
		
		// load the mc blocks
		loadMinecraftBlockImages();
		
		// build new image
		int newY = 0;
		for (int y = 0; y < height; ++y) {
			int newX = 0;
			for (int x = 0; x < width; ++x) {
				// get mc map id for color
				Color col = new Color(squareImg.getRGB(x, y), true);
				
				int mapId = 0;
				if (col.getAlpha() == 255) {
					// full color pixel
					mapId = findMapColor(col, minecraftMapColors).getBaseId();
				}
				
				// get block image for map id
				BufferedImage mapImage = minecraftBlocks.get(mapId);
				
				// draw block
				g.drawImage(mapImage, newX, newY, null);
				
				// build newX
				newX += 16;
				
				// draw seperator if not last pixel in line
				if (x < width - 1) {
					g.drawLine(newX, newY, newX, newY + 16);
					newX += 1;
				}
			}
			
			newY += 16;
			
			// draw line separator if not last pixel in height
			if (y < height - 1) {
				g.drawLine(0, newY, newW, newY);
				newY += 1;
			}
		}
		g.dispose();
		
		//save new file
		ImageIO.write(saveImg, "png", file);
	}
	
	public static void exportAsMap(BufferedImage image, File file, MinecraftMapColor[] minecraftMapColors) throws IOException {
		BufferedImage squareImg = getSquareImg(image);
		int width = squareImg.getWidth();
		int height = squareImg.getHeight();
		
		// write map
		Map<String, Tag> dataMap = new HashMap<String, Tag>();
		dataMap.put("scale", new ByteTag("scale", (byte)0)); // nonzoomed map
		dataMap.put("dimension", new ByteTag("dimension", (byte)0)); // normal world
		dataMap.put("height", new ShortTag("height", (short)128));
		dataMap.put("width", new ShortTag("width", (short)128));
		dataMap.put("xCenter", new IntTag("xCenter", 13371337)); // same as ImageToMap
		dataMap.put("zCenter", new IntTag("zCenter", -13371337)); // same as ImageToMap
		
		byte[] colors = new byte[width * height];
		// colorID = Colors[widthOffset + heightOffset * width], where (widthOffset==0, heightOffset==0) is left upper point.
		
		// firstly, get an x y array of the image as MC map color ID
		for (int y = 0; y < height; ++y) {
			for (int x = 0; x < width; ++x) {
				// get mc map id for color
				Color col = new Color(squareImg.getRGB(x, y), true);
				
				byte mapId = 0;
				if (col.getAlpha() == 255) {
					// full color pixel
					mapId = (byte)findMapColor(col, minecraftMapColors).getColorId();
				}
				
				colors[x + (y * width)] = mapId;
			}
		}

		dataMap.put("colors", new ByteArrayTag("colors", colors));
		
		CompoundTag data = new CompoundTag("data", dataMap);
		Map<String, Tag> masterMap = new HashMap<String, Tag>();
		masterMap.put("data", data);
		CompoundTag master = new CompoundTag("", masterMap);
		NBTOutputStream nos = new NBTOutputStream(new FileOutputStream(file));
		nos.writeTag(master);
		nos.close();
	}
	
	/**
	 * Returns a squared instance of the image, with the larger dimension being used for both
	 * For example, if width of sourceImage is larger than height, the returned image is of
	 * dimension width x width with the source image vertically centered between borders of transparency
	 * @param sourceImage
	 * @return
	 */
	private static BufferedImage getSquareImg(BufferedImage sourceImage) {
		int origWidth = sourceImage.getWidth();
		int origHeight = sourceImage.getHeight();
		
		int width = origWidth;
		int height = origHeight;
		int xPos = 0;
		int yPos = 0;
		if (origWidth > origHeight) {
			// new image is width x width
			height = origWidth;
			yPos = (height - origHeight) / 2;
		} else if (origHeight > origWidth) {
			// new image is height x height
			width = origHeight;
			xPos = (width - origWidth) / 2;
		}
		
		BufferedImage newImg = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
		Graphics2D g = newImg.createGraphics();
		
		// fill with transparency
		g.setColor(new Color(255, 255, 255, 0));
		g.fillRect(0, 0, width, height);
		
		// add image
		g.drawImage(sourceImage, xPos, yPos, null);
		g.dispose();
		return newImg;
	}
	
	private static void loadMinecraftBlockImages() throws IOException {
		if (minecraftBlocks != null) return;
		minecraftBlocks = new HashMap<Integer, BufferedImage>();
		
		// load images according to http://minecraft.gamepedia.com/Map_item_format (base id)
		// image files are already named accordingly, with x.png where x is the map format base ID
		
		// load images by package name so it works from jar too
		//ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
		String paramPath = "klaue/minecraftimagemap/resources/";
		
		for (int i = 0; i <= 35; ++i) {
			BufferedImage img = ImageIO.read(ClassLoader.getSystemResource(paramPath + i + ".png"));
            minecraftBlocks.put(i, img);
		}
	}
	
	private static MinecraftMapColor findMapColor(Color c, MinecraftMapColor[] mapColors) {
		for (MinecraftMapColor mapCol : mapColors) {
			if (mapCol.getColor().equals(c)) return mapCol;
		}
		
		// no mapping found, return transparent block as default
		return new MinecraftMapColor(new Color(255, 255, 255, 0), 1, 0);
	}
}
