package klaue.minecraftimagemap;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.filechooser.FileFilter;



public class MinecraftImageMap extends JFrame implements ActionListener {
	private static final long serialVersionUID = 3556941853034815633L;

	static Color[] minecraftBaseColors = new Color[] { new Color(127, 178, 56),
		new Color(247, 233, 163), new Color(167, 167, 167),
		new Color(255, 0, 0), new Color(160, 160, 255),
		new Color(167, 167, 167), new Color(0, 124, 0),
		new Color(255, 255, 255), new Color(164, 168, 184),
		new Color(183, 106, 47), new Color(112, 112, 112),
		new Color(64, 64, 255), new Color(104, 83, 50),
		new Color(255, 252, 245), new Color(216, 127, 51),
		new Color(178, 76, 216), new Color(102, 153, 216),
		new Color(229, 229, 51), new Color(127, 204, 25),
		new Color(242, 127, 165), new Color(76, 76, 76),
		new Color(153, 153, 153), new Color(76, 127, 153),
		new Color(127, 63, 178), new Color(51, 76, 178),
		new Color(102, 76, 51), new Color(102, 127, 51),
		new Color(153, 51, 51), new Color(25, 25, 25),
		new Color(250, 238, 77), new Color(92, 219, 213),
		new Color(74, 128, 255), new Color(0, 217, 58),
		new Color(21, 20, 31), new Color(112, 2, 0) };
	
	static MinecraftMapColor[] minecraftMapColors = new MinecraftMapColor[minecraftBaseColors.length * 4];
	
	static {
		// each base color has 4 map colors, see http://minecraft.gamepedia.com/Map_item_format#Map_colors
		// each new color is built by multiplying r, g and b by 180, 220, 255 and 135 and then dividing by 255
		int[] multipliers = new int[] {180, 220, 255, 135};
		int colIndex = 0;
		for (int i = 0; i < minecraftBaseColors.length; ++i) {
			double r = minecraftBaseColors[i].getRed();
			double g = minecraftBaseColors[i].getGreen();
			double b = minecraftBaseColors[i].getBlue();
			
			for (int j = 0; j < multipliers.length; ++j) {
				int newR = (int)((r * multipliers[j])/255);
				int newG = (int)((g * multipliers[j])/255);
				int newB = (int)((b * multipliers[j])/255);
				Color newColor = new Color(newR, newG, newB);
				
				// we left out the first, transparent block in the base colors, so base value (i) starts one later
				minecraftMapColors[colIndex + j] = new MinecraftMapColor(newColor, (i+1), ((i+1)*4) + j);
			}
			colIndex += 4;
		}
	}
	
	JPanel mainPanel = new JPanel();
	JPanel imagePanel = new JPanel();
	JPanel toolsPanel = new JPanel();
	JPanel radioPanel = new JPanel();
	JPanel exportPanel = new JPanel();
	
	JButton loadImage = new JButton("Load image");
	JButton exportMap = new JButton("Export as MC-Map");
	JButton exportMcImage = new JButton("Export as MC block image");
	JButton exportImage = new JButton("Export as PNG");
	
	JRadioButton allColors = new JRadioButton("All colors");
	JRadioButton baseColors = new JRadioButton("Only base colors");
	ButtonGroup colorRadios = new ButtonGroup();
	
	JRadioButton ditheringRadio = new JRadioButton("Dithering");
	JRadioButton noDitheringRadio = new JRadioButton("No dithering");
	ButtonGroup ditheringRadios = new ButtonGroup();
	
	BufferedImage originalResizedImage = null;
	BufferedImage indexedImage = null;
	
	/**
	 * sets up the GUI
	 */
	public MinecraftImageMap() {
		setTitle("Minecraft Image Map");
		setSize(550, 450);
		setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		setLocationRelativeTo(null);
		
		mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.Y_AXIS));
		mainPanel.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
		
		mainPanel.add(imagePanel);
		
		toolsPanel.setLayout(new BoxLayout(toolsPanel, BoxLayout.X_AXIS));
		toolsPanel.add(loadImage);
		toolsPanel.add(Box.createHorizontalStrut(10));
		
		radioPanel.setLayout(new BoxLayout(radioPanel, BoxLayout.Y_AXIS));
		
		colorRadios.add(allColors);
		colorRadios.add(baseColors);
		allColors.setSelected(true);

		ditheringRadios.add(ditheringRadio);
		ditheringRadios.add(noDitheringRadio);
		ditheringRadio.setSelected(true);
		
		radioPanel.add(allColors);
		radioPanel.add(baseColors);
		radioPanel.add(Box.createVerticalStrut(10));
		
		radioPanel.add(ditheringRadio);
		radioPanel.add(noDitheringRadio);
		toolsPanel.add(radioPanel);
		toolsPanel.add(Box.createHorizontalStrut(10));
		
		exportPanel.setLayout(new BoxLayout(exportPanel, BoxLayout.Y_AXIS));
		exportMap.setEnabled(false);
		exportMcImage.setEnabled(false);
		exportImage.setEnabled(false);
		exportPanel.add(exportMap);
		exportPanel.add(exportMcImage);
		exportPanel.add(exportImage);
		toolsPanel.add(exportPanel);
		
		toolsPanel.setMaximumSize(new Dimension(toolsPanel.getPreferredSize().width, toolsPanel.getMinimumSize().height));
		mainPanel.add(toolsPanel);
		
		this.add(mainPanel);
		
		// add actions
		loadImage.setActionCommand("LOAD");
		loadImage.addActionListener(this);
		ditheringRadio.setActionCommand("RADIO");
		noDitheringRadio.setActionCommand("RADIO");
		allColors.setActionCommand("RADIO");
		baseColors.setActionCommand("RADIO");
		ditheringRadio.addActionListener(this);
		noDitheringRadio.addActionListener(this);
		allColors.addActionListener(this);
		baseColors.addActionListener(this);
		exportMap.setActionCommand("EXPORTMAP");
		exportMcImage.setActionCommand("EXPORTBLOCKS");
		exportImage.setActionCommand("EXPORTPNG");
		exportMap.addActionListener(this);
		exportMcImage.addActionListener(this);
		exportImage.addActionListener(this);
		
		
		this.setVisible(true);
	}

	@Override
	public void actionPerformed(ActionEvent e) {
		if (e.getActionCommand().equals("LOAD")) {
			JFileChooser jfcImageChooser = new JFileChooser();
			jfcImageChooser.setFileFilter(new FileFilter() {
				public boolean accept(File arg0) {
					if (arg0.isDirectory()) return true;
					String fileName = arg0.getName();
					if (!fileName.contains(".")) return false;
					String fileExtension = fileName.substring(fileName.lastIndexOf(".") + 1);
					
					String[] extensions = ImageIO.getWriterFormatNames();
					for (String extension : extensions) {
						if (extension.equals(fileExtension)) return true;
					}
					return false;
				}
				public String getDescription() {
					return "Image files";
				}
			});
			int answer = jfcImageChooser.showOpenDialog(this);
			if (answer != JFileChooser.APPROVE_OPTION) return;
			File imageFile = jfcImageChooser.getSelectedFile();
			
			BufferedImage image = null;
			try {
				image = ImageIO.read(imageFile);
			} catch (IOException e1) {
				e1.printStackTrace();
				return;
			}
			
			originalResizedImage = scaleImage(128, 128, image, Image.SCALE_SMOOTH);
				
			reloadImage();
			
			exportMap.setEnabled(true);
			//exportMcImage.setEnabled(true);
			exportImage.setEnabled(true);
		} else if (e.getActionCommand().equals("RADIO")) {
			if (originalResizedImage == null) return;
			reloadImage();
		} else if (e.getActionCommand().startsWith("EXPORT")) {
			if (indexedImage == null) return;
			
			JFileChooser jfcSave = new JFileChooser();
			int answer = jfcSave.showSaveDialog(this);
			if (answer != JFileChooser.APPROVE_OPTION) return;
			
			String allowedExtension = ".png";
			if (e.getActionCommand().equals("EXPORTMAP")) {
				allowedExtension = ".dat";
			}
			
			
			File saveFile = jfcSave.getSelectedFile();
			if (!saveFile.getName().toLowerCase().endsWith(allowedExtension)) {
				saveFile = new File(saveFile.getAbsolutePath() + allowedExtension);
			}
			
			if (saveFile.exists()) {
				int reply = JOptionPane.showConfirmDialog(this, "The file allready exists, overwrite?", "File exists", JOptionPane.YES_NO_OPTION);
				if (reply != JOptionPane.YES_OPTION) return;
				saveFile.delete();
			}
			
			try {
				if (e.getActionCommand().equals("EXPORTPNG")) {
					Exporter.exportAsPNG(indexedImage, saveFile);
				} else if (e.getActionCommand().equals("EXPORTBLOCKS")) {
					Exporter.exportAsBlocks(indexedImage, saveFile, minecraftMapColors);
				} else if (e.getActionCommand().equals("EXPORTMAP")) {
					Exporter.exportAsMap(indexedImage, saveFile, minecraftMapColors);
				}
			} catch (IOException e1) {
				JOptionPane.showMessageDialog(this, "Sorry, unable to save, " + e1.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
			}
		} else if (e.getActionCommand().equals("EXPORTMAP")) {
			
		}
		
	}
	
	/**
	 * scales the image to the given size, keeping aspect ratio
	 * @param width
	 * @param height
	 * @param sourceImage
	 * @param hints Image.SCALE_SMOOTH or Image.SCALE_FAST
	 * @return a scaled copy of the given image
	 */
	private BufferedImage scaleImage(int width, int height, BufferedImage sourceImage, int hints) {
		int origWidth = sourceImage.getWidth();
		int origHeight = sourceImage.getHeight();
		
		// get real new dimensions
		int newWidth = width;
		int newHeight = height;
		if (origWidth < origHeight) {
			newWidth = (int)(origWidth * ((double)height / origHeight));
		} else if (origWidth > origHeight) {
			newHeight = (int)(origHeight * ((double)width / origWidth));
		}
		
		Image toolkitImage = sourceImage.getScaledInstance(newWidth, newHeight, hints);
		BufferedImage scaled = new BufferedImage(newWidth, newHeight, BufferedImage.TYPE_INT_ARGB);
		Graphics2D g = scaled.createGraphics();
		
		// set resized image
		g.drawImage(toolkitImage, 0, 0, null);
		g.dispose();
		return scaled;
	}
	
	private void reloadImage() {
		Color[] colors = null;
		if (baseColors.isSelected()) {
			colors = minecraftBaseColors;
		} else {
			// reform the minecraft map colors array to a normal color array
			colors = new Color[minecraftMapColors.length];
			for (int i = 0; i < minecraftMapColors.length; ++i) {
				colors[i] = minecraftMapColors[i].getColor();
			}
		}
		
		if (this.ditheringRadio.isSelected()) {
			this.indexedImage = ColorIndexer.floydSteinbergDithering(this.originalResizedImage, colors);
		} else {
			this.indexedImage = ColorIndexer.noDithering(this.originalResizedImage, colors);
		}
		
		// make a new image, scaled to 2x, to display
		BufferedImage displayImg = scaleImage(256, 256, this.indexedImage, Image.SCALE_FAST);
		this.imagePanel.removeAll();
		this.imagePanel.add(new JLabel(new ImageIcon(displayImg)));
		this.imagePanel.doLayout();

		if (baseColors.isSelected()) {
			exportMcImage.setEnabled(true);
		} else {
			exportMcImage.setEnabled(false);
		}
	}
	
	public static void main(String[] args) throws IOException {
		new MinecraftImageMap();
	}
}
